import jinja2
import typing
from bs4 import BeautifulSoup, Tag
from jinja2 import lexer, nodes
from jinja2.ext import Extension
from jinja2.parser import Parser
from jinja2.runtime import Context
from starlette.requests import HTTPConnection

from starlette_live.components import import_component


class LiveComponentTag(Extension):
    tags = {'live_component'}

    def parse(self, parser: Parser) -> typing.Union[nodes.Node, typing.List[nodes.Node]]:
        if parser.stream.current.value == 'live_component':
            return self.parse_live_component(parser)
        raise ValueError('Unsupported tag')

    def parse_live_component(self, parser: Parser) -> typing.Union[nodes.Node, typing.List[nodes.Node]]:
        lineno = parser.stream.current.lineno
        parser.stream.skip(1)

        args: list = []
        kwargs: list = [
            nodes.Keyword('_context', nodes.ContextReference()),
        ]

        kwargs_set = False
        while parser.stream.current.type != lexer.TOKEN_BLOCK_END:
            parser.stream.skip_if(lexer.TOKEN_COMMA)
            if parser.stream.current.type == 'name' and parser.stream.look().type == 'assign':
                key = parser.stream.current.value
                parser.stream.skip(2)
                value = parser.parse_expression()
                kwargs.append(nodes.Keyword(key, value, lineno=value.lineno))
                kwargs_set = True
            else:
                if kwargs_set:
                    parser.fail('Invalid argument syntax', parser.stream.current.lineno)
                args.append(parser.parse_expression())

        block_call = self.call_method('render_live_component', args, kwargs)
        return nodes.Output([nodes.MarkSafe(block_call)]).set_lineno(lineno)

    def render_live_component(self, name: str, **kwargs: typing.Any) -> str:
        context: Context = kwargs.pop('_context')
        environment: jinja2.Environment = context.environment
        conn: HTTPConnection = context.parent['conn']

        component_class = import_component(name)
        component = component_class(**kwargs)
        content = component.render(environment, conn)
        conn.state.components.register(component)

        soup = BeautifulSoup(content, features='html.parser')
        if not soup.contents:
            return content

        has_multiple_roots = len([child for child in soup.contents if isinstance(child, Tag)]) > 1
        if has_multiple_roots:
            raise ValueError(
                f'Component {component_class.__module__}.{component_class.__name__} (id="{component.id}") '
                f'has multiple root elements.'
            )

        soup.contents[0].attrs['live-component'] = component.id
        return str(soup)



import jinja2
import json
import logging
import typing
from starlette import status
from starlette.datastructures import FormData, Headers
from starlette.formparsers import FormParser
from starlette.requests import HTTPConnection
from starlette.responses import HTMLResponse
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket

from starlette_live.components import ComponentRegistry


class Message(typing.TypedDict):
    method: str
    params: dict[str, typing.Any] | FormData


class IncomingMessage(typing.TypedDict):
    method: str
    target: str
    params: str | dict[str, typing.Any]
    format: typing.Literal['json', 'form']


class LiveView:
    template = ''
    component_template = ''
    block = 'content'
    jinja_env: jinja2.Environment | None = None

    def __init__(self) -> None:
        self._component_registry = ComponentRegistry()

    async def state(self, conn: HTTPConnection) -> dict[str, typing.Any]:
        return {}

    async def dispatch(self, websocket: WebSocket, message: Message) -> None:
        pass

    def render(self, state: dict[str, typing.Any]) -> str:
        assert self.template
        assert self.jinja_env
        template = self.jinja_env.get_template(self.template)
        return template.render(state)

    async def update(self, websocket: WebSocket) -> None:
        state = await self._make_state(websocket)
        content = self.render(state)
        await self.send_command(websocket, 'render', {'content': content, 'target': 'root'})

    async def _make_state(self, conn: HTTPConnection) -> dict[str, typing.Any]:
        state = await self.state(conn)
        state['conn'] = conn
        return state

    async def _dispatch_http(self, conn: HTTPConnection) -> HTMLResponse:
        state = await self._make_state(conn)
        content = self.render(state)
        return HTMLResponse(content)

    async def send_command(self, websocket: WebSocket, action: str, params: dict[str, str | int]) -> None:
        await websocket.send_json({'method': action, 'params': params})

    async def _dispatch_websocket(self, websocket: WebSocket) -> None:
        await websocket.accept()

        # re-render page contents to initialize live components
        state = await self._make_state(websocket)
        self.render(state)

        # mount components
        await self.mount_components(websocket)

        while True:
            message = await websocket.receive()
            if message["type"] == "websocket.receive":
                payload: IncomingMessage = json.loads(message['text'])
                await self._handle_websocket_message(websocket, payload)

            elif message['type'] == 'websocket.disconnect':
                await self.unmount_components(websocket)
                break

    async def _handle_websocket_message(self, websocket: WebSocket, payload: IncomingMessage) -> None:
        method = payload['method']
        data_format = payload['format']
        params: dict[str, typing.Any] | FormData

        if data_format == 'json':
            assert isinstance(payload['params'], dict)
            params = payload['params']
        elif data_format == 'form':
            assert isinstance(payload['params'], str)
            params = await self._parse_form_data(payload['params'])
        else:
            raise TypeError('Unsupported payload params type.')

        assert isinstance(params, (dict, FormData))
        await self.dispatch(
            websocket,
            typing.cast(
                Message,
                {
                    'method': method,
                    'params': params,
                },
            ),
        )

    def _parse_json(self, data: str) -> typing.Any:
        return json.loads(data)

    async def _parse_form_data(self, data: str) -> FormData:
        async def stream() -> typing.AsyncGenerator[bytes, None]:
            yield data.encode()
            yield b''

        form_parser = FormParser(Headers(), stream())
        return await form_parser.parse()

    async def mount_components(self, websocket: WebSocket) -> None:
        try:
            for component in websocket.state.components:
                await component.mount(websocket)
        except Exception as ex:
            logging.exception(ex)
            await websocket.close(status.WS_1011_INTERNAL_ERROR)

    async def unmount_components(self, websocket: WebSocket) -> None:
        for component in websocket.state.components:
            try:
                await component.unmount(websocket)
            except Exception as ex:
                logging.exception(ex)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope.setdefault('state', {})
        scope['state']['components'] = self._component_registry

        if scope['type'] == 'http':
            request = HTTPConnection(scope, receive)
            response = await self._dispatch_http(request)
            await response(scope, receive, send)

        if scope['type'] == 'websocket':
            websocket = WebSocket(scope, receive=receive, send=send)
            await self._dispatch_websocket(websocket)



import typing
from starlette.datastructures import URLPath
from starlette.routing import BaseRoute, Match, NoMatchFound, compile_path, get_name, replace_params
from starlette.types import Receive, Scope, Send

from starlette_live.endpoints import LiveView


class LiveRoute(BaseRoute):
    def __init__(
        self,
        path: str,
        view_class: typing.Type[LiveView],
        name: typing.Optional[str] = None,
    ) -> None:
        assert path.startswith("/"), "Routed paths must start with '/'"
        self.path = path
        self.name = get_name(view_class) if name is None else name
        self.view_class = view_class
        self.methods = {'get', 'post'}
        self.path_regex, self.path_format, self.param_convertors = compile_path(path)

    def matches(self, scope: Scope) -> typing.Tuple[Match, Scope]:
        if scope["type"] in {"http", "websocket"}:
            match = self.path_regex.match(scope["path"])
            if match:
                matched_params = match.groupdict()
                for key, value in matched_params.items():
                    matched_params[key] = self.param_convertors[key].convert(value)
                path_params = dict(scope.get("path_params", {}))
                path_params.update(matched_params)
                child_scope = {"endpoint": self.view_class, "path_params": path_params}
                if scope['type'] == 'http' and self.methods and scope["method"] not in self.methods:
                    return Match.PARTIAL, child_scope
                else:
                    return Match.FULL, child_scope
        return Match.NONE, {}

    def url_path_for(self, name: str, **path_params: typing.Any) -> URLPath:
        seen_params = set(path_params.keys())
        expected_params = set(self.param_convertors.keys())

        if name != self.name or seen_params != expected_params:
            raise NoMatchFound(name, path_params)

        path, remaining_params = replace_params(self.path_format, self.param_convertors, path_params)
        assert not remaining_params
        return URLPath(path=path, protocol="http")

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        live_view = self.view_class()
        await live_view(scope, receive, send)



import importlib
import jinja2
import typing
from starlette.requests import HTTPConnection
from starlette.websockets import WebSocket


class Component:
    id: str = ''
    template: str = ''
    _counter: int = 0

    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        cls._counter += 1
        cls.id = cls.id or cls.__name__.lower()

    async def mount(self, websocket: WebSocket) -> None:
        pass

    def render(self, jinja_env: jinja2.Environment, conn: HTTPConnection) -> str:
        assert self.template
        template = jinja_env.get_template(self.template)
        return template.render({'this': self, 'conn': conn})


class ComponentRegistry:
    def __init__(self) -> None:
        self.components: dict[str, Component] = {}

    def register(self, component: Component) -> None:
        assert component.id not in self.components, f'Duplicate component ID: {component.id}'
        self.components[component.id] = component

    def __iter__(self) -> typing.Iterator[Component]:
        yield from self.components.values()


def import_component(name: str) -> typing.Type[Component]:
    module_path, _, component_name = name.rpartition('.')
    module = importlib.import_module(module_path)
    return getattr(module, component_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starlette_live']

package_data = \
{'': ['*'], 'starlette_live': ['statics/live/*']}

install_requires = \
['Jinja2>=3.1,<4.0', 'python-multipart>=0.0.5,<0.0.6', 'starlette>=0.20,<0.21']

setup_kwargs = {
    'name': 'starlette-live',
    'version': '0.1.0',
    'description': 'Live views for Starlette.',
    'long_description': '# OhMyAdmin\n\nAwesome admin panel for your business.\n\n![PyPI](https://img.shields.io/pypi/v/ohmyadmin)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/alex-oleshkevich/ohmyadmin/Lint%20and%20test)\n![GitHub](https://img.shields.io/github/license/alex-oleshkevich/ohmyadmin)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/ohmyadmin)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/ohmyadmin)\n![GitHub Release Date](https://img.shields.io/github/release-date/alex-oleshkevich/ohmyadmin)\n\n## Installation\n\nInstall `ohmyadmin` using PIP or poetry:\n\n```bash\npip install ohmyadmin\n# or\npoetry add ohmyadmin\n```\n\n## Features\n\n-   TODO\n\n## Quick start\n\nSee example application in `examples/` directory of this repository.\n',
    'author': 'Alex Oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/alex-oleshkevich/starlette_live',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
